# django imports
from django.db import IntegrityError
from django.contrib.auth.models import User
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist

# permissions imports
from permissions.models import ObjectPermission
from permissions.models import ObjectPermissionInheritanceBlock
from permissions.models import Permission

# Permission #################################################################

def grant_permission(permission, user_group, obj):
    """Adds passed permission to passed group and object. Returns True if the
    permission was able to be added, otherwise False.

    **Parameters:**

        permission
            The permission which should be granted. Either a permission
            object or the codename of a permission.
        user_group
            The user or group for which the permission should be granted.
        obj
            The content object for which the permission should be granted.
    """
    if not isinstance(permission, Permission):
        try:
            permission = Permission.objects.get(codename = permission)
        except Permission.DoesNotExist:
            return False

    ct = ContentType.objects.get_for_model(obj)
    if isinstance(user_group, Group):
        try:
            ObjectPermission.objects.get(group=user_group, content_type = ct, content_id=obj.id, permission=permission)
        except ObjectPermission.DoesNotExist:
            try:
                result = ObjectPermission.objects.create(group=user_group, content=obj, permission=permission)
            except IntegrityError:
                return False
    else:
        try:
            ObjectPermission.objects.get(user=user_group, content_type = ct, content_id=obj.id, permission=permission)
        except ObjectPermission.DoesNotExist:
            try:
                result = ObjectPermission.objects.create(user=user_group, content=obj, permission=permission)
            except IntegrityError:
                return False
    return True

def remove_permission(permission, user_group, obj):
    """Removes passed permission from passed group and object. Returns True if
    the permission has been removed.

    **Parameters:**

        permission
            The permission which should be removed. Either a permission object
            or the codename of a permission.
        user_group
            The user or group for which a permission should be removed.
        obj
            The content object for which a permission should be removed.
    """
    if not isinstance(permission, Permission):
        try:
            permission = Permission.objects.get(codename = permission)
        except Permission.DoesNotExist:
            return False

    ct = ContentType.objects.get_for_model(obj)

    if isinstance(user_group, Group):
        try:
            op = ObjectPermission.objects.get(group=user_group, content_type = ct, content_id=obj.id, permission = permission)
        except ObjectPermission.DoesNotExist:
            return False
    else:
        try:
            op = ObjectPermission.objects.get(user=user_group, content_type = ct, content_id=obj.id, permission = permission)
        except ObjectPermission.DoesNotExist:
            return False
    op.delete()
    return True

def has_permission(codename, user, obj=None, groups=[]):
    """Checks whether the passed user has passed permission for passed object.

    **Parameters:**

    codename
        The permission's codename which should be checked.
    user
        The user for which the permission should be checked.
    obj
        The object for which the permission should be checked.
    groups
        If given these groups will be assigned to the user temporarily before 
        the permissions are checked. If you don't know why this is need you
        can safely ignore it.
    """

    if obj is None:
        return False

    if user.is_superuser:
        return True

    user_groups = list(Group.objects.filter(user=user))
    user_groups.extend(groups)

    ct = ContentType.objects.get_for_model(obj)

    while obj is not None:
        p = ObjectPermission.objects.filter(
            content_type=ct, content_id=obj.id, user=user, permission__codename = codename)

        if p.exists():
            return True

        p = ObjectPermission.objects.filter(
            content_type=ct, content_id=obj.id, group__in=user_groups, permission__codename = codename)

        if p.exists():
            return True
        
        if is_inherited(codename, obj) == False:
            return False

        try:
            obj = obj.get_parent_for_permissions()
            ct = ContentType.objects.get_for_model(obj)
        except AttributeError:
            return False

    return False

# Inheritance ################################################################

def add_inheritance_block(permission, obj):
    """Adds an inheritance for the passed permission on the passed obj.

    **Parameters:**

        permission
            The permission for which an inheritance block should be added.
            Either a permission object or the codename of a permission.
        obj
            The content object for which an inheritance block should be added.
    """
    if not isinstance(permission, Permission):
        try:
            permission = Permission.objects.get(codename = permission)
        except Permission.DoesNotExist:
            return False

    ct = ContentType.objects.get_for_model(obj)
    try:
        ObjectPermissionInheritanceBlock.objects.get(content_type = ct, content_id=obj.id, permission=permission)
    except ObjectPermissionInheritanceBlock.DoesNotExist:
        try:
            result = ObjectPermissionInheritanceBlock.objects.create(content=obj, permission=permission)
        except IntegrityError:
            return False
    return True

def remove_inheritance_block(permission, obj):
    """Removes a inheritance block for the passed permission from the passed
    object.

    **Parameters:**

        permission
            The permission for which an inheritance block should be removed.
            Either a permission object or the codename of a permission.
        obj
            The content object for which an inheritance block should be added.
    """
    if not isinstance(permission, Permission):
        try:
            permission = Permission.objects.get(codename = permission)
        except Permission.DoesNotExist:
            return False

    ct = ContentType.objects.get_for_model(obj)
    try:
        opi = ObjectPermissionInheritanceBlock.objects.get(content_type = ct, content_id=obj.id, permission=permission)
    except ObjectPermissionInheritanceBlock.DoesNotExist:
        return False

    opi.delete()
    return True

def is_inherited(codename, obj):
    """Returns True if the passed permission is inherited for passed object.

    **Parameters:**

        codename
            The permission which should be checked. Must be the codename of
            the permission.
        obj
            The content object for which the permission should be checked.
    """
    ct = ContentType.objects.get_for_model(obj)
    try:
        ObjectPermissionInheritanceBlock.objects.get(
            content_type=ct, content_id=obj.id, permission__codename = codename)
    except ObjectDoesNotExist:
        return True
    else:
        return False

def get_group(name):
    """Returns the group with given or None if it doesn't exit.
    """
    try:
        return Group.objects.get(name=name)
    except Group.DoesNotExist:
        return None
    
# Registering ################################################################

def register_permission(name, codename):
    """Registers a permission to the framework. Returns the permission if the
    registration was successfully, otherwise False.

    **Parameters:**

        name
            The unique name of the permission. This is displayed to the
            customer.

        codename
            The unique codename of the permission. This is used internally to
            identify the permission.
    """
    try:
        p = Permission.objects.create(name=name, codename=codename)
    except IntegrityError:
        return False
    return p

def unregister_permission(codename):
    """Unregisters a permission from the framework

    **Parameters:**

        codename
            The unique codename of the permission.
    """
    try:
        permission = Permission.objects.get(codename=codename)
    except Permission.DoesNotExist:
        return False
    permission.delete()
    return True

def register_group(name):
    """Registers a group with passed name to the framework. Creates a Django
    default group. Returns the new group if the registration was successfully,
    otherwise False.

    **Parameters:**

        name
            The unique group name.
    """
    try:
        group = Group.objects.create(name=name)
    except IntegrityError:
        return False
    return group

def unregister_group(name):
    """Unregisters the group with passed name. This will remove a Django
    default group with passed name.

    **Parameters:**

        name
            The unique group name.
    """
    try:
        group = Group.objects.get(name=name)
    except Group.DoesNotExist:
        return False

    group.delete()
    return True


# django imports
from django.db import models
from django.contrib.auth.models import User
from django.contrib.auth.models import Group
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

class Permission(models.Model):
    """A permission which can be granted to users/groups and objects.

    **Attributes:**

        name
            The unique name of the permission. This is displayed to users.
        codename
            The unique codename of the permission. This is used internal to
            identify a permission.
    """
    name = models.CharField(_(u"Name"), max_length=100, unique=True)
    codename = models.CharField(_(u"Codename"), max_length=100, unique=True)

    def __unicode__(self):
        return "%s (%s)" % (self.name, self.codename)

class ObjectPermission(models.Model):
    """Grants permission for specific user/group and object.

    **Attributes:**

        group
            The group for which the permission is granted. Either this xor the
            user must be given.
        user
            The user for which the permission is granted. Either this xor the
            user must be given.
        permission
            The permission which is granted.
        content
            The object for which the permission is granted.
    """
    group = models.ForeignKey(Group, verbose_name=_(u"Group"), blank=True, null=True)
    user = models.ForeignKey(User, verbose_name=_(u"User"), blank=True, null=True)
    permission = models.ForeignKey(Permission, verbose_name=_(u"Permission"))

    content_type = models.ForeignKey(ContentType, verbose_name=_(u"Content type"))
    content_id = models.PositiveIntegerField(verbose_name=_(u"Content id"))
    content = generic.GenericForeignKey(ct_field="content_type", fk_field="content_id")

    def __unicode__(self):
        if self.group:
            principal = self.group
        else:
            principal = self.user

        return "%s / %s / %s - %s" % (self.permission.name, principal, self.content_type, self.content_id)

class ObjectPermissionInheritanceBlock(models.Model):
    """Blocks the inheritance for specific permission and object.

    **Attributes:**

        permission
            The permission for which inheritance is blocked.
        content
            The object for which the inheritance is blocked.
    """
    permission = models.ForeignKey(Permission, verbose_name=_(u"Permission"))

    content_type = models.ForeignKey(ContentType, verbose_name=_(u"Content type"))
    content_id = models.PositiveIntegerField(verbose_name=_(u"Content id"))
    content = generic.GenericForeignKey(ct_field="content_type", fk_field="content_id")

    def __unicode__(self):
        return "%s / %s - %s" % (self.permission, self.content_type, self.content_id)



# django imports
from django.contrib.flatpages.models import FlatPage
from django.contrib.auth.models import Group
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test.client import Client

# permissions imports
from permissions.models import Permission
import permissions.utils

class PermissionTestCase(TestCase):
    """
    """
    def setUp(self):
        """
        """
        self.group_1 = permissions.utils.register_group("Group 1")
        self.group_2 = permissions.utils.register_group("Group 2")

        self.user = User.objects.create(username="john")
        self.user.groups.add(self.group_1)
        self.user.groups.add(self.group_2)
        self.user.save()

        self.page_1 = FlatPage.objects.create(url="/page-1/", title="Page 1")
        self.page_2 = FlatPage.objects.create(url="/page-1/", title="Page 2")

        self.permission = permissions.utils.register_permission("View", "view")

    def test_has_permission_group(self):
        """
        """
        result = permissions.utils.has_permission("view", self.user, self.page_1)
        self.assertEqual(result, False)

        result = permissions.utils.grant_permission(self.permission, self.group_1, self.page_1)
        self.assertEqual(result, True)

        result = permissions.utils.has_permission("view", self.user, self.page_1)
        self.assertEqual(result, True)

        result = permissions.utils.remove_permission("view", self.group_1, self.page_1)
        self.assertEqual(result, True)

        result = permissions.utils.has_permission("view", self.user, self.page_1)
        self.assertEqual(result, False)

    def test_has_permission_owner(self):
        """
        """
        creator = User.objects.create(username="jane")

        result = permissions.utils.has_permission("view", creator, self.page_1)
        self.assertEqual(result, False)

        owner = permissions.utils.register_group("Owner")
        permissions.utils.grant_permission("view", owner, self.page_1)

        result = permissions.utils.has_permission("view", creator, self.page_1, [owner])
        self.assertEqual(result, True)

    def test_has_permission_user(self):
        """
        """
        result = permissions.utils.has_permission("view", self.user, self.page_1)
        self.assertEqual(result, False)

        result = permissions.utils.grant_permission(self.permission, self.user, self.page_1)
        self.assertEqual(result, True)

        result = permissions.utils.has_permission("view", self.user, self.page_1)
        self.assertEqual(result, True)

        result = permissions.utils.remove_permission("view", self.user, self.page_1)
        self.assertEqual(result, True)

        result = permissions.utils.has_permission("view", self.user, self.page_1)
        self.assertEqual(result, False)

    def test_ineritance(self):
        """
        """
        result = permissions.utils.is_inherited("view", self.page_1)
        self.assertEqual(result, True)

        permissions.utils.add_inheritance_block(self.permission, self.page_1)

        result = permissions.utils.is_inherited("view", self.page_1)
        self.assertEqual(result, False)

class RegistrationTestCase(TestCase):
    """Tests the registration of different components.
    """
    def test_group(self):
        """Tests registering/unregistering of a group.
        """
        # Register a group
        result = permissions.utils.register_group("Brights")
        self.failUnless(isinstance(result, Group))

        # It's there
        group = Group.objects.get(name="Brights")
        self.assertEqual(group.name, "Brights")

        # Trying to register another group with same name
        result = permissions.utils.register_group("Brights")
        self.assertEqual(result, False)

        group = Group.objects.get(name="Brights")
        self.assertEqual(group.name, "Brights")

        # Unregister the group
        result = permissions.utils.unregister_group("Brights")
        self.assertEqual(result, True)

        # It's not there anymore
        self.assertRaises(Group.DoesNotExist, Group.objects.get, name="Brights")

        # Trying to unregister the group again
        result = permissions.utils.unregister_group("Brights")
        self.assertEqual(result, False)

    def test_permission(self):
        """Tests registering/unregistering of a permission.
        """
        # Register a permission
        result = permissions.utils.register_permission("Change", "change")
        self.failUnless(isinstance(result, Permission))

        # Is it there?
        p = Permission.objects.get(codename="change")
        self.assertEqual(p.name, "Change")

        # Register a permission with the same codename
        result = permissions.utils.register_permission("Change2", "change")
        self.assertEqual(result, False)

        # Is it there?
        p = Permission.objects.get(codename="change")
        self.assertEqual(p.name, "Change")

        # Register a permission with the same name
        result = permissions.utils.register_permission("Change", "change2")
        self.assertEqual(result, False)

        # Is it there?
        p = Permission.objects.get(codename="change")
        self.assertEqual(p.name, "Change")

        # Unregister the permission
        result = permissions.utils.unregister_permission("change")
        self.assertEqual(result, True)

        # Is it not there anymore?
        self.assertRaises(Permission.DoesNotExist, Permission.objects.get, codename="change")

        # Unregister the permission again
        result = permissions.utils.unregister_permission("change")
        self.assertEqual(result, False)


# django imports
from django.core.handlers.wsgi import WSGIRequest
from django.contrib.auth.models import User
from django.contrib.sessions.backends.file import SessionStore
from django.test.client import Client

# Taken from "http://www.djangosnippets.org/snippets/963/"
class RequestFactory(Client):
    """
    Class that lets you create mock Request objects for use in testing.

    Usage:

    rf = RequestFactory()
    get_request = rf.get('/hello/')
    post_request = rf.post('/submit/', {'foo': 'bar'})

    This class re-uses the django.test.client.Client interface, docs here:
    http://www.djangoproject.com/documentation/testing/#the-test-client

    Once you have a request object you can pass it to any view function,
    just as if that view had been hooked up using a URLconf.

    """
    def request(self, **request):
        """
        Similar to parent class, but returns the request object as soon as it
        has created it.
        """
        environ = {
            'HTTP_COOKIE': self.cookies,
            'PATH_INFO': '/',
            'QUERY_STRING': '',
            'REQUEST_METHOD': 'GET',
            'SCRIPT_NAME': '',
            'SERVER_NAME': 'testserver',
            'SERVER_PORT': 80,
            'SERVER_PROTOCOL': 'HTTP/1.1',
        }
        environ.update(self.defaults)
        environ.update(request)
        return WSGIRequest(environ)

def create_request():
    """
    """
    rf = RequestFactory()
    request = rf.get('/')
    request.session = SessionStore()

    user = User()
    user.is_superuser = True
    user.save()
    request.user = user

    return request